/*
 * jdatasrc.c
 *
 * Copyright (C) 1994, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains decompression data source routines for the case of
 * reading JPEG data from a file (or any stdio stream).  While these routines
 * are sufficient for most applications, some will want to use a different
 * source manager.
 * IMPORTANT: we assume that fread() will correctly transcribe an array of
 * JOCTETs from 8-bit-wide elements on external storage.  If char is wider
 * than 8 bits on your machine, you may need to do some tweaking.
 */


/* this is not a core library module, so it doesn't define JPEG_INTERNALS */
#include "jinclude.h"
#include "jpeglib.h"
#include "jerror.h"

typedef struct
{
    struct jpeg_source_mgr pub;   /* public fields */

    JOCTET * buffer;              /* image buffer */
    unsigned int buffer_size;     /* image buffer size */
} mem_source_mgr;


typedef mem_source_mgr * mem_src_ptr;



/* Expanded data source object for stdio input */

typedef struct
{
    struct jpeg_source_mgr pub;	/* public fields */

    unsigned char *infile;		/* source stream */
    JOCTET * buffer;		/* start of buffer */
    boolean start_of_file;	/* have we gotten any data yet? */
} my_source_mgr;

typedef my_source_mgr * my_src_ptr;

#define INPUT_BUF_SIZE  4096	/* choose an efficiently fread'able size */


/*
 * Initialize source --- called by jpeg_read_header
 * before any data is actually read.
 */

METHODDEF void
init_source (j_decompress_ptr cinfo)
{
    mem_src_ptr src = (mem_src_ptr) cinfo->src;

    /* Initialize public members */

    /* See the notes in mem_dest.c for init_destination */

    /* next_input_byte is set to point to the beginning of the source
    ** image buffer
    */
    src->pub.next_input_byte = src->buffer;

    /* bytes_in_buffer is set to the maximum size of the input buffer.
    ** It is assumed that even if the JPEG compressed image data is smaller
    ** than buffer_size, the JFIF library will see the EOI tag and stop
    ** decomression before reaching the end of the buffer.
    */
    src->pub.bytes_in_buffer = src->buffer_size;
}


/*
 * Fill the input buffer --- called whenever buffer is emptied.
 *
 * In typical applications, this should read fresh data into the buffer
 * (ignoring the current state of next_input_byte & bytes_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been reloaded.  It is not necessary to
 * fill the buffer entirely, only to obtain at least one more byte.
 *
 * There is no such thing as an EOF return.  If the end of the file has been
 * reached, the routine has a choice of ERREXIT() or inserting fake data into
 * the buffer.  In most cases, generating a warning message and inserting a
 * fake EOI marker is the best course of action --- this will allow the
 * decompressor to output however much of the image is there.  However,
 * the resulting error message is misleading if the real problem is an empty
 * input file, so we handle that case specially.
 *
 * In applications that need to be able to suspend compression due to input
 * not being available yet, a FALSE return indicates that no more data can be
 * obtained right now, but more may be forthcoming later.  In this situation,
 * the decompressor will return to its caller (with an indication of the
 * number of scanlines it has read, if any).  The application should resume
 * decompression after it has loaded more data into the input buffer.  Note
 * that there are substantial restrictions on the use of suspension --- see
 * the documentation.
 *
 * When suspending, the decompressor will back up to a convenient restart point
 * (typically the start of the current MCU). next_input_byte & bytes_in_buffer
 * indicate where the restart point will be if the current call returns FALSE.
 * Data beyond this point must be rescanned after resumption, so move it to
 * the front of the buffer rather than discarding it.
 */

METHODDEF boolean
fill_input_buffer (j_decompress_ptr cinfo)
{
    my_src_ptr src = (my_src_ptr) cinfo->src;

    memcpy( src->buffer, src->infile, INPUT_BUF_SIZE );

    src->infile += INPUT_BUF_SIZE;

    src->pub.next_input_byte = src->buffer;
    src->pub.bytes_in_buffer = INPUT_BUF_SIZE;
    src->start_of_file = FALSE;

    return TRUE;
}

// from http://72.14.253.104/search?q=cache:NGXMiNIfwbsJ:www.tac.dk/software/viz/src/imageio/jmem_src.c+fill_input_buffer&hl=en&ct=clnk&cd=9&gl=us&lr=lang_en
// instead of filling an input buffer, this will EOF the stream and write out an error
METHODDEF boolean
fill_input_buffer_mem (j_decompress_ptr cinfo)
{
    mem_src_ptr src = (mem_src_ptr) cinfo->src;

    //  syserror("jpeg6: fill_input_buffer: should not be called.\n");

    /* Print a warning message */
    WARNMS(cinfo, JWRN_JPEG_EOF);

    /* Insert a fake EOI marker */
    src->buffer[0] = (JOCTET) 0xFF;
    src->buffer[1] = (JOCTET) JPEG_EOI;
    src->pub.bytes_in_buffer = 2;
    src->pub.next_input_byte = src->buffer;

    return TRUE;
}

/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * Writers of suspendable-input applications must note that skip_input_data
 * is not granted the right to give a suspension return.  If the skip extends
 * beyond the data currently in the buffer, the buffer can be marked empty so
 * that the next read will cause a fill_input_buffer call that can suspend.
 * Arranging for additional bytes to be discarded before reloading the input
 * buffer is the application writer's problem.
 */

METHODDEF void
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
    my_src_ptr src = (my_src_ptr) cinfo->src;

    /* Just a dumb implementation for now.  Could use fseek() except
     * it doesn't work on pipes.  Not clear that being smart is worth
     * any trouble anyway --- large skips are infrequent.
     */
    if (num_bytes > 0)
    {
        while (num_bytes > (long) src->pub.bytes_in_buffer)
        {
            num_bytes -= (long) src->pub.bytes_in_buffer;
            (void) fill_input_buffer_mem(cinfo);
            /* note we assume that fill_input_buffer will never return FALSE,
             * so suspension need not be handled.
             */
        }
        src->pub.next_input_byte += (size_t) num_bytes;
        src->pub.bytes_in_buffer -= (size_t) num_bytes;
    }
}


/*
 * An additional method that can be provided by data source modules is the
 * resync_to_restart method for error recovery in the presence of RST markers.
 * For the moment, this source module just uses the default resync method
 * provided by the JPEG library.  That method assumes that no backtracking
 * is possible.
 */


/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

METHODDEF void
term_source (j_decompress_ptr cinfo)
{
    /* no work necessary here */
}


/*
 * Prepare for input from a stdio stream.
 * The caller must have already opened the stream, and is responsible
 * for closing it after finishing decompression.
 */

GLOBAL void
jpeg_stdio_src (j_decompress_ptr cinfo, unsigned char *infile)
{
    my_src_ptr src;

    /* The source object and input buffer are made permanent so that a series
     * of JPEG images can be read from the same file by calling jpeg_stdio_src
     * only before the first one.  (If we discarded the buffer at the end of
     * one image, we'd likely lose the start of the next one.)
     * This makes it unsafe to use this manager and a different source
     * manager serially with the same JPEG object.  Caveat programmer.
     */
    if (cinfo->src == NULL)  	/* first time for this JPEG object? */
    {
        cinfo->src = (struct jpeg_source_mgr *)
                     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                             SIZEOF(my_source_mgr));
        src = (my_src_ptr) cinfo->src;
        src->buffer = (JOCTET *)
                      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                              INPUT_BUF_SIZE * SIZEOF(JOCTET));
    }

    src = (my_src_ptr) cinfo->src;
    src->pub.init_source = init_source;
    src->pub.fill_input_buffer = fill_input_buffer;
    src->pub.skip_input_data = skip_input_data;
    src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
    src->pub.term_source = term_source;
    src->infile = infile;
    src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
    src->pub.next_input_byte = NULL; /* until buffer loaded */
}



/* jpeg_memory_src: Prepare for input from memory.
**
** The user is responsible for having already allocated the image
** buffer, filling it before calling this decompression module, and will
** be responsible for freeing that buffer when it is no longer being
** used.
**
** cinfo: (in/out) pointer to decompression object
** jfif_buffer: (in) decompression image buffer
** buf_size: (in) maximum size of the jfif_buffer, in bytes.
**
** Note that buf_size is the MAXIMUM size of the buffer, not the size of
** the image.  This is so the same jfif_buffer can be re-filled with a new
** JPEG compressed image, and the same mem_source object can be re-used
** to decompress multiple objects.  However, read the comments below
** concerning the caveat about re-use.
**
*/

GLOBAL void
jpeg_memory_src (j_decompress_ptr cinfo, const unsigned char *jfif_buffer, int buf_size)
{
    mem_src_ptr src;

    /* The source object is made permanent so that a series
     * of JPEG images can be read from the same jfif_buffer by calling
     * jpeg_memory_src only before the first one.  (If we discarded the
     * buffer at the end of one image, we'd likely lose the start of
     * the next one.)
     *
     * This makes it unsafe to use this manager and a different source
     * manager serially with the same JPEG object.  Caveat programmer.
     */
    if (cinfo->src == NULL)       /* first time for this JPEG object? */
    {
        cinfo->src = (struct jpeg_source_mgr *)
                     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                             ((size_t) sizeof(mem_source_mgr)));
    }

    src = (mem_src_ptr) cinfo->src;  /* for casting */

    /* Initialize method pointers */
    src->pub.init_source = init_source;
    src->pub.fill_input_buffer = fill_input_buffer_mem;
    src->pub.skip_input_data = skip_input_data;
    src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
    src->pub.term_source = term_source;


    /* Initialize private members */
    src->buffer = (JOCTET*)jfif_buffer;
    src->buffer_size = buf_size;
    //  src->pub.next_input_byte = src->buffer; /* until buffer loaded */
}
